%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%     Vortex Island Estimation of Distribution Algorithm          %
%       Dr. Yoan Martnez-Lpez, Camaguey University, Cuba        %
%       Eng. Miguel Bethencourt, Camaguey University, Cuba        %
%       Dt. Julio Madera, Camaguey University, Cuba               % 
%       Dr. Ansel Rodriguez, CICESE, Mexico                       %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Ref:
% Do?an, B., & lmez, T. (2015). A new metaheuristic for numerical function optimization: Vortex Search algorithm. Information Sciences, 293, 125-145.
% Madera, J., Alba, E., & Ochoa, A. (2006). A parallel island model for estimation of distribution algorithms. In Towards a new evolutionary computation (pp. 159-186). Springer, Berlin, Heidelberg.
% Martnez-Lpez, Y., Madera, J., Rodrguez-Gonzlez, A. Y., & Barigye, S. (2019). Cellular estimation Gaussian algorithm for continuous domain. Journal of Intelligent & Fuzzy Systems, 36(5), 4957-4967.
% Martnez-Lpez, Y., Rodrguez-Gonzlez, A. Y., Quintana, J. M., Moya, A., Morgado, B., & Mayedo, M. B. (2019, July). CUMDANCauchy-C1: a cellular EDA designed to solve the energy resource management problem under uncertainty. In Proceedings of the Genetic and Evolutionary Computation Conference Companion (pp. 13-14).
% Martnez-Lpez, Y., Rodrguez-Gonzlez, A. Y., Quintana, J. M., Mayedo, M. B., Moya, A., & Santiago, O. M. (2020, July). Applying some EDAs and hybrid variants to the ERM problem under uncertainty. In Proceedings of the 2020 Genetic and Evolutionary Computation Conference Companion (pp. 1-2).
% Martnez-Lpez, Y., Rodrguez-Gonzlez, A. Y., Madera, J., Mayedo, M. B., & Lezama, F. (2021). Cellular estimation of distribution algorithm designed to solve the energy resource management problem under uncertainty. Engineering Applications of Artificial Intelligence, 101, 104231.
% Rodrguez-Gonzlez, A. Y., Lezama, F., Martnez-Lpez, Y., Madera, J., Soares, J., & Vale, Z. (2022). WCCI/GECCO 2020 Competition on Evolutionary Computation in the Energy Domain: An overview from the winner perspective. Applied Soft Computing, 109162.
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% email: cybervalient@gmail.com, betamiguel00@gmail.com
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


function [Fit_and_p,FVr_bestmemit, fitMaxVector] = ...
    VIEDA(VIEDAParameters,caseStudyData,otherParameters,low_habitat_limit,up_habitat_limit)

    %-----This is just for notational convenience and to keep the code uncluttered.--------
    I_NP         = VIEDAParameters.I_NP;
    I_D          = numel(up_habitat_limit); 
    
    
    I_itermax    = VIEDAParameters.I_itermax;
    FVr_minbound = low_habitat_limit;
    FVr_maxbound = up_habitat_limit;
    
    
    Sub_Pop              = VIEDAParameters.Sub_Pop;
    Sub_Pop_Size         = VIEDAParameters.Sub_Pop_Size;
    Elitist_Selection    = VIEDAParameters.Elitist_Selection;
    radius               = VIEDAParameters.radius;
    
    
    
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    fnc          =  otherParameters.fnc;
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %-----Check input variables---------------------------------------------
    if (I_NP < 5)
        I_NP=5;
        fprintf(1,' I_NP increased to minimal value 5\n');
    end
    if (I_itermax <= 0)
        I_itermax = 200;
        fprintf(1,'I_itermax should be > 0; set to default value 200\n');
    end
    
    %-----Initialize population and some arrays-------------------------------
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    % pre-allocation of loop variables
    fitMaxVector       =  nan(1,I_itermax);
    fitSolutionVector  =  nan(I_itermax,I_D);
    fitWorstVector     =  nan(1,I_itermax);
    fitAvgVector       =  nan(1,I_itermax);
    
    
    % limit iterations by threshold
    gen = 1; %iterations
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    %----FM_pop is a matrix of size I_NPx(I_D+1). It will be initialized------
    %----with random values between the min and max values of the-------------
    %----parameters-----------------------------------------------------------
    % FLC modification - vectorization
    minPositionsMatrix = repmat(FVr_minbound,I_NP,1);
    maxPositionsMatrix = repmat(FVr_maxbound,I_NP,1);
    
    % generate initial population.
    rand('state',otherParameters.iRuns) %Guarantee same initial population
    FM_pop = genpop(I_NP,I_D,minPositionsMatrix,maxPositionsMatrix);
    %FM_pop = genpop(I_NP,I_D,minPositionsMatrix,(minPositionsMatrix+maxPositionsMatrix)/2);
    %FM_middle = genpop(I_NP,I_D,minPositionsMatrix,(minPositionsMatrix+maxPositionsMatrix)/2);
    
    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    %------Evaluate the best member after initialization----------------------
    [S_val, ~]=feval(fnc,FM_pop,caseStudyData,otherParameters);

    
    %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    [S_best_val, I_best_index] = min(S_val); 
    FVr_bestmemit = FM_pop(I_best_index,:); 
    Bindex = I_best_index;
    
    avgV   = mean(S_val);
    worstV = max(S_val);
    
    fitMaxVector(1,gen)               = S_best_val;
    fitSolutionVector(gen, 1 : I_D)   = FVr_bestmemit;
    fitWorstVector(1,gen)             = worstV;
    fitAvgVector(1,gen)               = avgV;
    
    bestValue = S_best_val;
    %bestmemit = FVr_bestmemit;
    %------DE-Minimization---------------------------------------------
    %------FM_popold is the population which has to compete. It is--------
    %------static through one iteration. FM_pop is the newly--------------
    %------emerging population.----------------------------------------
    
    FM_newpop = FM_pop;
    %r = 0.2; 
    range = FVr_maxbound - FVr_minbound;
    
    while gen < I_itermax 
    
    
        SelectedIndividualIndexes = getSelectedIndividualIndexes(S_val,Elitist_Selection);
            
    
        for i = 1:Sub_Pop
           
             r = radius/10^(i); 
             old_Mu = FVr_minbound + range/(i+1);
            
           % [Mu,Sigma] = getEstimadedModel2(FM_pop,SelectedIndividualIndexes,old_Mu,r,range);
            [Mu,Sigma] =   getEstimadedModel3(FM_pop, SelectedIndividualIndexes, old_Mu,FVr_minbound,FVr_maxbound,r,range);
    
            SubPopIndividualIdexes = GetSubPopIndividualIdexes(i, Sub_Pop_Size);
         
            
    
            NewSubPop = sampleNewIndividualsFromModelholdBoudary(Sub_Pop_Size,Mu,Sigma);
              
            
            
            
            for j = 1: Sub_Pop_Size
                FM_newpop(SubPopIndividualIdexes(j),:) = NewSubPop(j,:);
            end
            
           %r = range/i;  
           %r = r - 0.005;
        end
        
      
%% Boundary Control

    
        FM_pop = getUnboudedPop(FM_newpop,FVr_minbound,FVr_maxbound);    
    
        %Evaluation of new Pop
        [S_val, ~]=feval(fnc,FM_pop,caseStudyData, otherParameters);

        
        
%% Update Best Results

    
        [S_best_val,I_best_index] = min(S_val);
        
        
        if(S_best_val < bestValue)
          FVr_bestmemit = FM_pop(I_best_index,:);
          Bindex = I_best_index;
          bestValue = S_best_val;
        end 
        
        S_val( Bindex) =  bestValue;
        FM_pop(Bindex ,:) = FVr_bestmemit ;
        
        fprintf('Best Worst Avg Fitness value: %f %f %f %f\n',bestValue,S_best_val,avgV,worstV);
        fprintf('Generation: %d\n',gen);
        
        gen = gen + 1;
        
        fitMaxVector(1,gen)               =  bestValue; %S_best_val
        fitSolutionVector(gen, 1 : I_D)   = FVr_bestmemit;
        fitWorstVector(1,gen)             = worstV;
        fitAvgVector(1,gen)               = avgV;
       
    end %---end while 
        Fit_and_p = [fitMaxVector(1,gen) 0];
    
%% Save xls Results

%         xlswrite('data.xlsx',FM_pop);
%         xlswrite('solutions.xlsx',fitSolutionVector);
end


%% 
% 
%%
function unboudedIndiv = rebound(unboudedIndiv,FVr_minbound,FVr_maxbound)

    [idx] = find(~(FVr_minbound <= unboudedIndiv));
    unboudedIndiv(idx) = FVr_minbound(idx);
    [idx] = find(~(unboudedIndiv <=FVr_maxbound()));
    unboudedIndiv(idx) = FVr_minbound(idx);

end

function unboudedPop = getUnboudedPop(FM_pop,FVr_minbound,FVr_maxbound)

    [hs,ws] = size(FM_pop);
    unboudedPop = nan(hs,ws);
    
    for i = 1: hs
    unboudedPop(i,:) = rebound(FM_pop(i,:),FVr_minbound,FVr_maxbound);
    end
        
end



function SubPopIndividualIdexes = GetSubPopIndividualIdexes(index,Sub_Pop_Size)
    
    SubPopIndividualIdexes = nan(1,Sub_Pop_Size,'single');
    base = (index - 1) * Sub_Pop_Size;
    
    for index = 1:Sub_Pop_Size
        SubPopIndividualIdexes(index) = base + index;
    end
    
end




function SelectedIndividualIndexes = getSelectedIndividualIndexes(S_val,Elitist_Selection)

    [~,I] = sort(S_val);       
    SelectedIndividualIndexes = I(1:Elitist_Selection);
 
end



function [Mu,Sigma] = getEstimadedModel(FM_pop,SelectedIndividualIndexes)

    parents = FM_pop(SelectedIndividualIndexes, :);
    
    Mu = mean(parents);
    
    Sigma = std(parents);
end


function parents = getEstimadedModel1(FM_pop,SelectedIndividualIndexes,old_Mu,FVr_minbound,FVr_maxbound,r,range)
  parents = FM_pop(SelectedIndividualIndexes, :);
  [nind,all] = size(parents);
  
 for n = 1: nind
     
  for i = 1: all
        if range(i) == 0
            continue;  
        end
        
        if parents(n,i) <= FVr_minbound(i) 
          parents(n,i) = parents(n,i) + (FVr_maxbound(i) - FVr_minbound(i))/2;
          continue;
        end
        
        if parents(n,i) <= old_Mu(i)
          parents(n,i) = parents(n,i) - r;       
           %Mu(i) = Mu(i) + r(i); 
           continue;
        end
               
        if parents(n,i) > old_Mu(i)
           parents(n,i) = parents(n,i) + r;   
          %Mu(i) = Mu(i)  - r(i); 
          continue;
        end

   end
  
 end
 
end

function [Mu,Sigma] = getEstimadedModel2(FM_pop, SelectedIndividualIndexes, old_Mu,r,range)

    parents = FM_pop(SelectedIndividualIndexes, :);        
    
    Mu = mean(parents); 
    
    Sigma = std(parents);
    
    
    [~,all] = size(FM_pop);

    for i = 1: all
        if range(i) == 0
            continue;  
        end
        
        if Mu(i) <= old_Mu(i)
           Mu(i) = Mu(i) - r;       
          % Mu(i) = Mu(i) + r(i); 
          continue;
        end
        
        if Mu(i) > old_Mu(i)
           Mu(i) = Mu(i) + r;   
          %Mu(i) = Mu(i)  - r(i); 
          continue;
        end

    end

end

function [Mu,Sigma] = getEstimadedModel3(FM_pop, SelectedIndividualIndexes, old_Mu,FVr_minbound,FVr_maxbound,r,range)

    parents = FM_pop(SelectedIndividualIndexes, :);        
    
    Mu = mean(parents); 
    
    Sigma = std(parents);
    
    
    [~,all] = size(FM_pop);

    for i = 1: all
        if range(i) == 0
            continue;  
        end
        
         if Mu(i) <= FVr_minbound(i) 
            Mu(i) = Mu(i) + Sigma(i);
          continue;
         end
        
         if Mu(i) >= (FVr_maxbound(i) + FVr_minbound(i))/8 %8 (18492) 4(18490) 2(18495)
           Mu(i) = Mu(i) - r;
          continue;
         end
         
        if Mu(i) <= old_Mu(i)
           Mu(i) = Mu(i) - r;       
           % Mu(i) = Mu(i) -  (range(i))*0.1; 
          continue;
        end
        
        if Mu(i) > old_Mu(i)
           Mu(i) = Mu(i) + r;   
           %Mu(i) = Mu(i)  + (range(i))*0.1;  
          continue;
        end

    end

end


function NewSubPop = sampleNewIndividualsFromModelholdBoudary(Sub_Pop_Size,Mu,Sigma)

    [~,ws] = size(Mu);
    NewSubPop = nan (Sub_Pop_Size,ws);              
    
    for j = 1: ws
        NewSubPop(:,j) = normrnd(Mu(j), Sigma(j), Sub_Pop_Size, 1);     
    end
    
end


% VECTORIZED THE CODE INSTEAD OF USING FOR
function pop=genpop(a,b,lowMatrix,upMatrix)
    pop=unifrnd(lowMatrix,upMatrix,a,b);
end